from skl2onnx import to_onnx
from skl2onnx.common.data_types import (
    BooleanTensorType,
    DoubleTensorType,
    FloatTensorType,
    Int32TensorType,
    Int64TensorType,
    StringTensorType,
)


def _str_to_type(content):
    content = content.lower().strip()
    if 'float32' == content:
        return FloatTensorType
    if 'float64' == content:
        return DoubleTensorType
    if 'int64' == content:
        return Int64TensorType
    if 'int' == content:
        return Int32TensorType
    if 'bool' == content:
        return BooleanTensorType
    if 'str' == content:
        return StringTensorType

    raise ValueError


def parse_skl_input(shape, dtype):
    type_cls = _str_to_type(dtype)
    return type_cls(shape)


def skl_convert_onnx(model, output_path, inputs_type, verbose=False):
    onx = to_onnx(model, initial_types=inputs_type, options={id(model): {'zipmap': False}})
    with open(output_path, "wb") as f:
        f.write(onx.SerializeToString())



import logging
from typing import OrderedDict

import tensorflow as tf
import tf2onnx
import torch
from onnxruntime.transformers import optimizer
from onnxruntime.transformers.fusion_options import FusionOptions


def transformers_convert_pytorch(
    model: torch.nn.Module,
    output_path: str,
    inputs_pytorch: OrderedDict[str, torch.Tensor],
    opset_version: int = 12,
    verbose: bool = False,
) -> None:
    """Convert an pytorch tansformer model to onnx.

    This model conversion is specific for transformers.

    Parameters
    ----------
    model: torch.nn.Module
        the pytorch model to convert to.
    output_path: str
        the onnx output filepath
    inputs_pytorch: OrderedDict[str, torch.Tensor]
        the model inputs
    opset_version: int
        the onnx op version to use. Default is 12.
    verbose: bool
        show detailed logging. Defaul is False.
    """
    dynamic_axis = {}
    for k in inputs_pytorch.keys():
        dynamic_axis[k] = {0: "batch_size", 1: "sequence"}
    dynamic_axis["output"] = {0: "batch_size"}
    with torch.no_grad():
        torch.onnx.export(
            model,
            args=tuple(inputs_pytorch.values()),
            f=output_path,
            opset_version=opset_version,
            do_constant_folding=True,
            input_names=list(inputs_pytorch.keys()),
            output_names=["output"],
            dynamic_axes=dynamic_axis,
            verbose=verbose,
        )


def transformers_convert_tf(
    model: tf.keras.Model,
    output_path: str,
    inputs_tf: OrderedDict[str, tf.Tensor],
    opset_version: int = 12,
    verbose: bool = False,
) -> None:
    """Convert an tensorflow transformers model to onnx.

    This model conversion is specific for transformers.

    Parameters
    ----------
    model: torch.nn.Module
        the tensorflow model to convert to.
    output_path: str
        the onnx output filepath
    inputs_tf: OrderedDict[str, torch.Tensor]
        the model inputs
    opset_version: int
        the onnx op version to use. Default is 12.
    verbose: bool
        show detailed logging. Defaul is False.
    """
    spec = tuple([tf.TensorSpec((None,) + v.shape[1:], v.dtype, name=k) for k, v in inputs_tf.items()])
    tf2onnx.convert.from_keras(model, input_signature=spec, opset=opset_version, output_path=output_path)


def transformers_optimize_onnx(
    onnx_path: str, output_path: str, model_type: str, use_cuda: bool, num_heads: int = 0, hidden_size: int = 0
) -> None:
    """Transformer model optimization.

    This apply custom optimization for transformer model type (encoder-only, decoder-only and encoder-decoder).

    Parameters
    ----------
    onnx_path: str
        onnx model path.
    output_path: str
        output onnx model path.
    model_type:str
        transformer model type. One of [bert, bart, gpt2].
    use_cuda: bool
        optimize for cuda.
    """
    logging.info(f"Optimizing for model type: {model_type}")
    optimization_options = FusionOptions(model_type)
    optimization_options.enable_gelu_approximation = True
    optimized_model = optimizer.optimize_model(
        input=onnx_path,
        model_type=model_type,
        use_gpu=use_cuda,
        opt_level=1,
        num_heads=num_heads,
        hidden_size=hidden_size,
        optimization_options=optimization_options,
        only_onnxruntime=True,
    )

    logging.info(f"optimizations applied: {optimized_model.get_fused_operator_statistics()}")
    optimized_model.save_model_to_file(output_path)



import multiprocessing
from typing import List

from onnxruntime import GraphOptimizationLevel, InferenceSession, SessionOptions
from onnxruntime.quantization import QuantType, quantize_dynamic


def generic_optimize_onnx(onnx_path: str, output_path: str) -> None:
    """Apply dynamic quantization to any onnx model.

    This can be used to any model, and also to second optimization step.

    Parameters
    ----------
    onnx_path: str
        path of onnx model to be optimized.
    output_path: str
        path of optimized onnx model.
    """
    quantize_dynamic(model_input=onnx_path, model_output=output_path, weight_type=QuantType.QInt8)


def create_model_for_provider(path: str, provider_to_use: str) -> InferenceSession:
    """Create onnx model for desired provider.

    This is used to run onnx models as standalone.

    Parameters
    ----------
    path: str
        onnx model path.
    provider_to_use: str
        execution provider, eg.: CPUExecutionProvider.

    Returns
    ----------
    InferenceSession
        the onnx model for inference.
    """
    options = SessionOptions()
    options.graph_optimization_level = GraphOptimizationLevel.ORT_ENABLE_ALL

    if provider_to_use == "CPUExecutionProvider":
        options.intra_op_num_threads = multiprocessing.cpu_count()

    providers: List[str] = [provider_to_use]
    return InferenceSession(path, options, providers=providers)



from typing import OrderedDict

import torch


def torch_convert_onnx(
    model: torch.nn.Module,
    output_path: str,
    inputs_pytorch: OrderedDict[str, torch.Tensor],
    opset_version: int = 12,
    verbose: bool = False,
) -> None:
    """Convert an pytorch model to onnx.

    This model conversion is specific for torch.

    Parameters
    ----------
    model: torch.nn.Module
        the pytorch model to convert to.
    output_path: str
        the onnx output filepath
    inputs_pytorch: OrderedDict[str, torch.Tensor]
        the model inputs
    opset_version: int
        the onnx op version to use. Default is 12.
    verbose: bool
        show detailed logging. Defaul is False.
    """
    dynamic_axis = {}
    for k in inputs_pytorch.keys():
        dynamic_axis[k] = {0: "batch_size"}
    dynamic_axis["output"] = {0: "batch_size"}
    with torch.no_grad():
        torch.onnx.export(
            model,
            args=tuple(inputs_pytorch.values()),
            f=output_path,
            opset_version=opset_version,
            do_constant_folding=True,
            input_names=list(inputs_pytorch.keys()),
            output_names=["output"],
            dynamic_axes=dynamic_axis,
            verbose=verbose,
        )



import numpy as np
from scipy.special import softmax


class OnnxPipeline:
    """Transformers pipeline compatible with onnx model

    This is used to create an standalone execution.

    Parameters
    ----------
    model: OnnxModel
        The onnx model to get the logits.
    tokenizer: Tokenizer
        Transformers tokenizer.
    intents:
        Convert tokens to words.
    """

    def __init__(self, model, tokenizer, intents=None):
        self.model = model
        self.tokenizer = tokenizer
        self.intents = intents

    def __call__(self, query):
        model_inputs = self.tokenizer(query, return_tensors="pt")
        inputs_onnx = {k: v.cpu().detach().numpy() for k, v in model_inputs.items()}
        logits = self.model.run(None, inputs_onnx)[0][0, :]
        probs = softmax(logits)
        pred_idx = np.argmax(probs).item()
        return [{"label": self.intents.int2str(pred_idx), "score": probs[pred_idx]}]



import logging
import re
from argparse import Namespace
from collections import OrderedDict
from typing import OrderedDict as OrderedDictType
from typing import Tuple

import numpy as np
import tensorflow as tf
import torch
import unidecode


def slugify(text):
    """Make words into an single word lower case word separable by '_'.

    This is used to construct endpoints from models name.

    Parameters
    ----------
    text: str
        Words to be slugify.

    Returns
    ----------
    str
        Words as one word separable by '_'.
    """
    text = unidecode.unidecode(text).lower()
    return re.sub(r'[\W_]+', '_', text)


def setup_logging(level: int = logging.INFO) -> None:
    """Setup the logging to desired level and format the message.

    This filter the logging by the level and format the message.

    Parameters
    ----------
    level: int
        The logging level.
    """
    logging.basicConfig(
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=level,
    )


def get_provider(args: Namespace) -> str:
    """Get provider by common argparse parameters.

    This return the execution ort provide: 'CUDAExecutionProvider' or  'CPUExecutionProvider'.

    Parameters
    ----------
    args: Namespace
        The argparse namespace.

    Returns
    ----------
    str
        'CUDAExecutionProvider' or  'CPUExecutionProvider'.
    """
    if args.cuda:
        assert torch.cuda.is_available(), "CUDA is not available. Please check your CUDA installation"
        return "CUDAExecutionProvider"

    return "CPUExecutionProvider"


def parse_transformer_torch_input(
    seq_len: int, batch_size: int, include_token_ids: bool, use_cuda: bool = False
) -> Tuple[OrderedDictType[str, torch.Tensor], OrderedDictType[str, np.ndarray]]:
    """Get transformers model input as torch and onnx.

    This is used to create torch and onnx model input configuration.

    Parameters
    ----------
    seq_len: int
        Transformer model sequence length.
    include_token_ids: bool
        Check if need return the token ids also.
    use_cuda: bool
        Place the tensors on cuda device

    Returns
    ----------
    Tuple[Tuple[Dict[str, Tensor]], Tuple[Dict[str, Tensor]]]
        A tuple of same shape for torch and onnx.
    """
    shape = (batch_size, seq_len)
    inputs_pytorch: OrderedDict[str, torch.Tensor] = OrderedDict()
    inputs_pytorch["input_ids"] = torch.randint(high=100, size=shape, dtype=torch.long)
    if include_token_ids:
        inputs_pytorch["token_type_ids"] = torch.ones(size=shape, dtype=torch.long)
    inputs_pytorch["attention_mask"] = torch.ones(size=shape, dtype=torch.long)
    inputs_onnx: OrderedDict[str, np.ndarray] = OrderedDict(
        {k: np.ascontiguousarray(v.detach().cpu().numpy()) for k, v in inputs_pytorch.items()}
    )

    if use_cuda:
        inputs_pytorch = OrderedDict({k: v.cuda() for k, v in inputs_pytorch.items()})

    return inputs_pytorch, inputs_onnx


def parse_torch_input(
    shape: Tuple[int, ...], batch_size: int, use_cuda: bool = False
) -> Tuple[OrderedDictType[str, torch.Tensor], OrderedDictType[str, np.ndarray]]:
    """Get model input as torch and onnx.

    This is used to create torch and onnx model input configuration.

    Parameters
    ----------
    seq_len: int
        Transformer model sequence length.
    include_token_ids: bool
        Check if need return the token ids also.
    use_cuda: bool
        Place the tensors on cuda device

    Returns
    ----------
    Tuple[Tuple[Dict[str, Tensor]], Tuple[Dict[str, Tensor]]]
        A tuple of same shape for torch and onnx.
    """
    shape = (batch_size,) + shape
    inputs_pytorch: OrderedDict[str, torch.Tensor] = OrderedDict()
    inputs_pytorch["input"] = torch.randint(high=100, size=shape, dtype=torch.long).type(torch.float)
    inputs_onnx: OrderedDict[str, np.ndarray] = OrderedDict(
        {k: np.ascontiguousarray(v.detach().cpu().numpy()) for k, v in inputs_pytorch.items()}
    )

    if use_cuda:
        inputs_pytorch = OrderedDict({k: v.cuda() for k, v in inputs_pytorch.items()})

    return inputs_pytorch, inputs_onnx


def parse_transformer_tf_input(
    seq_len: int, batch_size: int, include_token_ids: bool
) -> Tuple[OrderedDictType[str, tf.Tensor], OrderedDictType[str, np.ndarray]]:
    """Get transformers model input as tf and onnx.

    This is used to create tf and onnx model input configuration.

    Parameters
    ----------
    seq_len: int
        Transformer model sequence length.
    include_token_ids: bool
        Check if need return the token ids also.

    Returns
    ----------
    Tuple[Tuple[Dict[str, Tensor]], Tuple[Dict[str, Tensor]]]
        A tuple of same shape for tf and onnx.
    """
    shape = (batch_size, seq_len)
    inputs_tf: OrderedDict[str, tf.Tensor] = OrderedDict()
    inputs_tf["input_ids"] = tf.random.uniform(shape=shape, maxval=100, dtype=tf.int64)
    if include_token_ids:
        inputs_tf["token_type_ids"] = tf.ones(shape=shape, dtype=tf.int64)
    inputs_tf["attention_mask"] = tf.ones(shape=shape, dtype=tf.int64)
    inputs_onnx: OrderedDict[str, np.ndarray] = OrderedDict(
        {k: np.ascontiguousarray(v.numpy()) for k, v in inputs_tf.items()}
    )
    return inputs_tf, inputs_onnx


def parse_tf_input(
    shape: Tuple[int], batch_size: int
) -> Tuple[OrderedDictType[str, tf.Tensor], OrderedDictType[str, np.ndarray]]:
    """Get model input as tensorflow and onnx.

    This is used to create tensorflow and onnx model input configuration.

    Parameters
    ----------
    seq_len: int
        Transformer model sequence length.
    include_token_ids: bool
        Check if need return the token ids also.

    Returns
    ----------
    Tuple[Tuple[Dict[str, Tensor]], Tuple[Dict[str, Tensor]]]
        A tuple of same shape for tensorflow and onnx.
    """
    new_shape: Tuple[int, ...] = (batch_size,) + shape
    inputs_tf: OrderedDict[str, tf.Tensor] = OrderedDict()
    inputs_tf["input"] = tf.random.uniform(shape=new_shape, maxval=100, dtype=tf.float32)
    inputs_onnx: OrderedDict[str, np.ndarray] = OrderedDict(
        {k: np.ascontiguousarray(v.detach().cpu().numpy()) for k, v in inputs_tf.items()}
    )
    return inputs_tf, inputs_onnx



from pathlib import Path
from time import perf_counter
from typing import Callable

import numpy as np
from datasets import Dataset, load_metric


class PerformanceBenchmark:
    def __init__(self, pipeline: Callable, dataset: Dataset, name: str = "baseline"):
        self.pipeline = pipeline
        self.dataset = dataset
        self.name = name

    def compute_accuracy(self):
        pass

    def compute_size(self):
        pass

    def time_pipeline(self):
        pass

    def run_benchmark(self):
        metrics = {"name": self.name}
        metrics["size"] = self.compute_size()
        metrics["latency"] = self.time_pipeline()
        metrics["accuracy"] = self.compute_accuracy()
        return metrics


class LMPerformanceBenchmark(PerformanceBenchmark):
    def compute_accuracy(self):
        accuracy_score = load_metric("accuracy")
        intents = self.dataset.features["intent"]

        preds, labels = [], []
        for example in self.dataset:
            pred = self.pipeline(example["text"])[0]["label"]
            label = example["intent"]
            preds.append(intents.str2int(pred))
            labels.append(label)
        accuracy = accuracy_score.compute(predictions=preds, references=labels)
        print(f"Accuracy on test set - {accuracy['accuracy']:.3f}")
        return accuracy

    def time_pipeline(self):
        model_inputs = next(iter(self.dataset))["text"]
        latencies = []
        # Warmup
        for _ in range(10):
            _ = self.pipeline(model_inputs)
        # Timed run
        for _ in range(100):
            start_time = perf_counter()
            _ = self.pipeline(model_inputs)
            latency = perf_counter() - start_time
            latencies.append(latency)
        # Compute run statistics
        time_avg_ms = 1000 * np.mean(latencies)
        time_std_ms = 1000 * np.std(latencies)
        print(f"Average latency (ms) - {time_avg_ms:.2f} +\\- {time_std_ms:.2f}")
        return {"time_avg_ms": time_avg_ms, "time_std_ms": time_std_ms}


class OnnxPerformanceBenchmark(LMPerformanceBenchmark):
    def __init__(
        self,
        pipeline: Callable,
        dataset: Dataset,
        model_path: Path,
        name: str = "baseline",
    ):
        super().__init__(pipeline, dataset, name)
        self.model_path = model_path

    def compute_size(self):
        size_mb = Path(self.model_path).stat().st_size / (1024 * 1024)
        print(f"Model size (MB) - {size_mb:.2f}")
        return {"size_mb": size_mb}



# import tensorflow as tf

# from keras2onnx import convert_keras, save_model


# def tf_convert_onnx(
#     model: tf.keras.Model,
#     model_name: str,
#     output_path: str,
#     opset_version: int = 12,
#     verbose: bool = False,
# ) -> None:
#     """Convert an tensorflow tansformer model to onnx.

#     This model conversion is specific for transformers.

#     Parameters
#     ----------
#     model: torch.nn.Module
#         the tensorflow model to convert to.
#     output_path: str
#         the onnx output filepath
#     inputs_tf: OrderedDict[str, torch.Tensor]
#         the model inputs
#     opset_version: int
#         the onnx op version to use. Default is 12.
#     verbose: bool
#         show detailed logging. Defaul is False.
#     """
#     onnx_model = convert_keras(model, model_name, target_opset=opset_version)
#     save_model(onnx_model, output_path)



import os
import shutil
from enum import Enum
from pathlib import Path
from typing import List

from quick_deploy.utils import slugify


class TritonIOTypeConf(Enum):
    STRING = 'TYPE_STRING'
    FP32 = 'TYPE_FP32'
    INT64 = 'TYPE_INT64'

    @classmethod
    def from_str(cls, content):
        content = content.lower().strip()
        if 'string' == content:
            return TritonIOTypeConf.STRING
        if 'float32' == content:
            return TritonIOTypeConf.FP32
        if 'int64' == content:
            return TritonIOTypeConf.INT64

        raise ValueError


class TritonIOConf:
    name: str
    data_type: str
    dims: List[int]

    def __init__(self, name: str, data_type: TritonIOTypeConf, dims: List[int]):
        self.name = name
        self.data_type = data_type.value
        self.dims = dims


class TritonModelConf:
    def __init__(
        self,
        workind_directory: str,
        model_name: str,
        batch_size: int,
        nb_instance: int,
        use_cuda: bool,
        model_inputs: List[TritonIOConf],
        model_outputs: List[TritonIOConf],
    ):
        self.model_name = model_name
        self.folder_name = slugify(model_name)
        self.batch_size = batch_size
        assert nb_instance > 0, f"nb_instance=={nb_instance}: nb model instances should be positive"
        self.nb_instance = nb_instance
        self.include_token_type = True
        self.workind_directory = workind_directory
        self.input_type = "TYPE_INT64"
        self.inference_platform = "onnxruntime_onnx"
        self.kind = "KIND_GPU" if use_cuda else "KIND_CPU"
        self.model_inputs = model_inputs
        self.model_outputs = model_outputs

    def _get_inputs(self):
        inputs_conf = []
        for _input in self.model_inputs:
            input_conf = f"""{{
        name: "{_input.name}"
        data_type: {_input.data_type}
        dims: {_input.dims}
    }}"""
            inputs_conf.append(input_conf)
        return ",\n    ".join(inputs_conf)

    def _get_outputs(self):
        outputs_conf: List[str] = []
        for output in self.model_outputs:
            output_conf = f"""{{
        name: "{output.name}"
        data_type: {output.data_type}
        dims: {output.dims}
    }}"""
            outputs_conf.append(output_conf)
        return ",\n    ".join(outputs_conf)

    def _instance_group(self):
        return f"""
instance_group [
    {{
      count: {self.nb_instance}
      kind: {self.kind}
    }}
]
""".strip()

    def get_conf(self) -> str:
        return f"""
name: "{self.folder_name}"
max_batch_size: {self.batch_size}
platform: "{self.inference_platform}"
default_model_filename: "model.bin"

input [
    {self._get_inputs()}
]

output [
    {self._get_outputs()}
]

{self._instance_group()}
""".strip()

    def write(self, model_path: str):
        """Create the models files for Triton.

        This create model structure for Triton create the models endpoints
        and setup the ensemble of tokenizer+model using an custom python
        request handler.

        Parameters
        ----------
        model_path: str
            Model full path
        """
        wd_path = Path(self.workind_directory)
        wd_path.mkdir(parents=True, exist_ok=True)

        current_folder = wd_path.joinpath(self.folder_name)
        current_folder.mkdir(exist_ok=True)

        conf = self.get_conf()
        current_folder.joinpath("config.pbtxt").write_text(conf)
        version_folder = current_folder.joinpath("1")
        version_folder.mkdir(exist_ok=True)

        model_folder_path = wd_path.joinpath(self.folder_name).joinpath("1")
        shutil.copy(model_path, os.path.join(model_folder_path, "model.bin"))



from os.path import expanduser
import argparse
import logging
import pickle
from pathlib import Path

import numpy as np
import yaml

from quick_deploy.triton_template import TritonIOConf, TritonIOTypeConf, TritonModelConf

from quick_deploy.utils import (
    get_provider,
    parse_torch_input,
    parse_transformer_tf_input,
    parse_transformer_torch_input,
    setup_logging,
)

from quick_deploy.backend.common import create_model_for_provider, generic_optimize_onnx


def main_transformers(args):
    """Command to parse transformers models"""
    from transformers import pipeline

    from quick_deploy.backend.transformers_ort import (
        transformers_convert_pytorch,
        transformers_convert_tf,
        transformers_optimize_onnx,
    )

    onnx_model_path = Path(f"{expanduser(args.workdir)}/transformer_{args.name}.onnx").as_posix()
    onnx_optim_model_path = Path(f"{expanduser(args.workdir)}/transformer_{args.name}.optim.onnx").as_posix()

    provider_to_use = get_provider(args)

    pipe = pipeline(args.pipeline, model=args.model, tokenizer=args.tokenizer, framework=args.framework)
    pipe_tokenizer = pipe.tokenizer
    pipe_model = pipe.model

    if args.cuda:
        pipe_model.cuda()

    input_names = pipe_tokenizer.model_input_names
    include_token_ids = "token_type_ids" in input_names

    output_np: np.ndarray = None
    if "pt" == args.framework:
        import torch

        pipe_model.eval()
        inputs_pytorch, inputs_onnx = parse_transformer_torch_input(
            batch_size=1, seq_len=args.seq_len, include_token_ids=include_token_ids, use_cuda=args.cuda
        )

        with torch.inference_mode():
            output = pipe_model(**inputs_pytorch)
            output = output.logits
            output_np = output.detach().cpu().numpy()

        transformers_convert_pytorch(
            model=pipe_model, output_path=onnx_model_path, inputs_pytorch=inputs_pytorch, verbose=args.verbose
        )
    elif "tf" == args.framework:
        inputs_tf, inputs_onnx = parse_transformer_tf_input(
            batch_size=1, seq_len=args.seq_len, include_token_ids=include_token_ids
        )

        output = pipe_model(inputs_tf)
        output = output.logits
        output_np = output.numpy()

        transformers_convert_tf(
            model=pipe_model, output_path=onnx_model_path, inputs_tf=inputs_tf, verbose=args.verbose
        )
    else:
        raise ValueError

    del pipe_model

    if args.atol is not None:
        onnx_model = create_model_for_provider(path=onnx_model_path, provider_to_use=provider_to_use)
        output_onnx = onnx_model.run(None, inputs_onnx)
        assert np.allclose(a=output_onnx, b=output_np, atol=args.atol)

    to_optimize: str = onnx_model_path
    if args.model_type is not None:
        transformers_optimize_onnx(
            onnx_path=onnx_model_path,
            output_path=onnx_optim_model_path,
            model_type=args.model_type,
            use_cuda=args.cuda,
            num_heads=args.num_heads,
            hidden_size=args.hidden_size,
        )
        to_optimize = onnx_optim_model_path

    generic_optimize_onnx(onnx_path=to_optimize, output_path=onnx_optim_model_path)

    if args.atol is not None:
        onnx_model = create_model_for_provider(path=onnx_optim_model_path, provider_to_use=provider_to_use)
        output_onnx_optimised = onnx_model.run(None, inputs_onnx)
        assert np.allclose(a=output_onnx_optimised, b=output_np, atol=args.atol)

    input_ids = TritonIOConf(name='input_ids', data_type=TritonIOTypeConf.INT64, dims=[-1, -1])

    token_type_ids = TritonIOConf(name='token_type_ids', data_type=TritonIOTypeConf.INT64, dims=[-1, -1])

    attention_mask = TritonIOConf(name='attention_mask', data_type=TritonIOTypeConf.INT64, dims=[-1, -1])

    output_shape = [-1 for _ in output_np.shape[:-1]]
    output_shape = output_shape + [output_np.shape[-1]]
    output = TritonIOConf(name='output', data_type=TritonIOTypeConf.FP32, dims=output_shape)

    model_input = [input_ids, attention_mask]
    model_output = [output]

    if include_token_ids:
        model_input.insert(1, token_type_ids)

    model_conf = TritonModelConf(
        workind_directory=args.output,
        model_name=args.name,
        batch_size=0,
        nb_instance=1,
        use_cuda=args.cuda,
        model_inputs=model_input,
        model_outputs=model_output,
    )
    model_conf.write(onnx_optim_model_path)


def main_torch(args):
    """Command to parse torch models"""
    import torch

    from quick_deploy.backend.torch_ort import torch_convert_onnx

    torch_model = torch.load(args.model)
    if args.cuda:
        torch_model.cuda()

    torch_model.eval()
    provider_to_use = get_provider(args)

    with open(args.file, "r") as stream:
        io_conf = yaml.safe_load(stream)

    assert "IOSchema" == io_conf['kind']

    model_input = []
    for m_input in io_conf['inputs']:
        t_conf = TritonIOConf(
            name=m_input['name'], data_type=TritonIOTypeConf.from_str(m_input['dtype']), dims=[-1] + m_input['shape']
        )
        model_input.append(t_conf)

    model_output = []
    for m_output in io_conf['outputs']:
        t_conf = TritonIOConf(
            name=m_output['name'],
            data_type=TritonIOTypeConf.from_str(m_output['dtype']),
            dims=[-1] + m_output['shape'],
        )
        model_output.append(t_conf)

    input_shape = tuple(model_input[-1].dims[1:])
    inputs_pytorch, inputs_onnx = parse_torch_input(shape=input_shape, batch_size=1, use_cuda=args.cuda)

    with torch.inference_mode():
        output = torch_model(inputs_pytorch['input'])
        output_np: np.ndarray = output.detach().cpu().numpy()

    onnx_model_path = Path(f"{expanduser(args.workdir)}/torch_{args.name}.onnx").as_posix()
    onnx_optim_model_path = Path(f"{expanduser(args.workdir)}/torch_{args.name}.optim.onnx").as_posix()

    torch_convert_onnx(
        model=torch_model, output_path=onnx_model_path, inputs_pytorch=inputs_pytorch, verbose=args.verbose
    )
    del torch_model

    if args.atol is not None:
        onnx_model = create_model_for_provider(path=onnx_model_path, provider_to_use=provider_to_use)
        output_onnx = onnx_model.run(None, inputs_onnx)
        assert np.allclose(a=output_onnx, b=output_np, atol=args.atol)

    if args.no_quant:
        onnx_optim_model_path = onnx_model_path
    else:
        generic_optimize_onnx(onnx_path=onnx_model_path, output_path=onnx_optim_model_path)

    model_conf = TritonModelConf(
        workind_directory=args.output,
        model_name=args.name,
        batch_size=0,
        nb_instance=1,
        use_cuda=args.cuda,
        model_inputs=model_input,
        model_outputs=model_output,
    )
    model_conf.write(onnx_optim_model_path)


def main_skl(args):
    """Command to parse sklearn models."""
    from quick_deploy.backend.skl_ort import parse_skl_input, skl_convert_onnx

    with open(args.model, "rb") as p_file:
        model = pickle.load(p_file)

    with open(args.file, "r") as stream:
        io_conf = yaml.safe_load(stream)

    assert "IOSchema" == io_conf['kind']

    model_input = []
    initial_type = []
    for m_input in io_conf['inputs']:
        t_conf = TritonIOConf(
            name=m_input['name'], data_type=TritonIOTypeConf.from_str(m_input['dtype']), dims=[-1] + m_input['shape']
        )
        model_input.append(t_conf)
        initial_type.append((m_input['name'], parse_skl_input([None] + m_input['shape'], m_input['dtype'])))

    model_output = []
    for m_output in io_conf['outputs']:
        t_conf = TritonIOConf(
            name=m_output['name'], data_type=TritonIOTypeConf.from_str(m_output['dtype']), dims=m_output['shape']
        )
        model_output.append(t_conf)

    onnx_model_path = Path(f"{expanduser(args.workdir)}/skl_{args.name}.onnx").as_posix()

    skl_convert_onnx(model=model, output_path=onnx_model_path, inputs_type=initial_type, verbose=args.verbose)

    model_conf = TritonModelConf(
        workind_directory=args.output,
        model_name=args.name,
        batch_size=0,
        nb_instance=1,
        use_cuda=False,
        model_inputs=model_input,
        model_outputs=model_output,
    )
    model_conf.write(onnx_model_path)


def main_xgb(args):
    """Command to parse xgboost models."""
    from quick_deploy.backend.xgb_ort import parse_xgb_input, xgb_convert_onnx

    with open(args.model, "rb") as p_file:
        model = pickle.load(p_file)

    with open(args.file, "r") as stream:
        io_conf = yaml.safe_load(stream)

    assert "IOSchema" == io_conf['kind']
    onnx_model_path = Path(f"{expanduser(args.workdir)}/xgb_{args.name}.onnx").as_posix()

    model_input = []
    initial_type = []
    for m_input in io_conf['inputs']:
        t_conf = TritonIOConf(
            name=m_input['name'], data_type=TritonIOTypeConf.from_str(m_input['dtype']), dims=[-1] + m_input['shape']
        )
        model_input.append(t_conf)
        initial_type.append((m_input['name'], parse_xgb_input([None] + m_input['shape'], m_input['dtype'])))

    model_output = []
    for m_output in io_conf['outputs']:
        t_conf = TritonIOConf(
            name=m_output['name'], data_type=TritonIOTypeConf.from_str(m_output['dtype']), dims=m_output['shape']
        )
        model_output.append(t_conf)

    xgb_convert_onnx(model=model, output_path=onnx_model_path, inputs_type=initial_type, verbose=args.verbose)

    model_conf = TritonModelConf(
        workind_directory=args.output,
        model_name=args.name,
        batch_size=0,
        nb_instance=1,
        use_cuda=False,
        model_inputs=model_input,
        model_outputs=model_output,
    )
    model_conf.write(onnx_model_path)


def default_args(parser):
    """Parse common arguments between commands."""
    parser.add_argument("-n", "--name", required=True, help="model name")
    parser.add_argument("-m", "--model", required=True, help="model path")
    parser.add_argument("-o", "--output", required=True, help="path used to export models")
    parser.add_argument("-w", "--workdir", default="~/.quick_deploy/", help="model path")
    parser.add_argument("--nb-instances", default=1, help="# of model instances", type=int)
    parser.add_argument("--cuda", action="store_true", help="use cuda optimization")
    parser.add_argument("-v", "--verbose", action="store_true", help="display detailed information")
    parser.add_argument("--atol", default=None, help="test outputs when convert", type=float)


def transformers_args(parser_tra):
    """Parse transformers arguments."""
    parser_tra.add_argument("-t", "--tokenizer", help="tokenizer path")
    parser_tra.add_argument(
        "--model-type",
        help="custom optimization for transformer model type. One of [bert, bart, gpt2]",
        choices=["bert", "bart", "gpt2"],
    )
    parser_tra.add_argument(
        "--framework",
        type=str,
        choices=["pt", "tf"],
        default="pt",
        help="Framework for loading the model",
    )
    parser_tra.add_argument("--num-heads", default=0, help="number of heads (not needed for bert)", type=int)
    parser_tra.add_argument("--hidden-size", default=0, help="the weights size (not needed for bert)", type=int)
    parser_tra.add_argument(
        "-p",
        "--pipeline",
        required=True,
        help="pipeline task, eg: 'text-classification'",
    )
    parser_tra.add_argument("--seq-len", default=16, help="sequence length to optimize", type=int)
    parser_tra.set_defaults(func=main_transformers)


def torch_args(parser_torch):
    """Parse torch arguments."""
    parser_torch.add_argument("-f", "--file", required=True, help="model IO configuration.")
    parser_torch.add_argument("--no-quant", action="store_true", help="avoid quant optimization")
    parser_torch.set_defaults(func=main_torch)


def main():
    """Entry-point function."""
    parser = argparse.ArgumentParser(
        description="Optimize and deploy machine learning models fast as possible!",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(help="transformers help", dest="transformers, torch")
    subparsers.required = True

    # transformers arguments and func binding
    parser_tra = subparsers.add_parser("transformers")
    default_args(parser_tra)
    transformers_args(parser_tra)

    # torch arguments and func binding
    parser_torch = subparsers.add_parser("torch")
    default_args(parser_torch)
    torch_args(parser_torch)

    # sklearn arguments and func binding
    parser_skl = subparsers.add_parser("sklearn")
    default_args(parser_skl)
    parser_skl.add_argument("-f", "--file", required=True, help="model IO configuration.")
    parser_skl.set_defaults(func=main_skl)

    # xgboost arguments and func binding
    parser_xgb = subparsers.add_parser("xgboost")
    default_args(parser_xgb)
    parser_xgb.add_argument("-f", "--file", required=True, help="model IO configuration.")
    parser_xgb.set_defaults(func=main_xgb)

    args = parser.parse_args()
    setup_logging(level=logging.INFO if args.verbose else logging.WARNING)

    Path(expanduser(args.workdir)).mkdir(parents=True, exist_ok=True)
    Path(expanduser(args.output)).mkdir(parents=True, exist_ok=True)
    args.func(args)


if __name__ == "__main__":
    main()



# import onnxmltools
from onnxmltools.convert.xgboost.operator_converters.XGBoost import (  # noqa
    convert_xgboost,
)

from skl2onnx import convert_sklearn, update_registered_converter

from skl2onnx.common.data_types import (
    BooleanTensorType,
    DoubleTensorType,
    FloatTensorType,
    Int32TensorType,
    Int64TensorType,
    StringTensorType,
)
from skl2onnx.common.shape_calculator import (  # noqa
    calculate_linear_classifier_output_shapes,
)
from xgboost import XGBClassifier

# import onnxmltools.convert.common.data_types


def _str_to_type(content):
    content = content.lower().strip()
    if 'float32' == content:
        return FloatTensorType
    if 'float64' == content:
        return DoubleTensorType
    if 'int64' == content:
        return Int64TensorType
    if 'int' == content:
        return Int32TensorType
    if 'bool' == content:
        return BooleanTensorType
    if 'str' == content:
        return StringTensorType

    raise ValueError


def parse_xgb_input(shape, dtype):
    type_cls = _str_to_type(dtype)
    return type_cls(shape)


def xgb_convert_onnx(model, output_path, inputs_type, verbose=False):
    # onx = to_onnx(model, initial_types=inputs_type, options={id(model): {'zipmap': False}})

    update_registered_converter(
        XGBClassifier,
        'XGBoostXGBClassifier',
        calculate_linear_classifier_output_shapes,
        convert_xgboost,
        options={'nocl': [True, False], 'zipmap': [True, False, 'columns']},
    )
    # update_registered_converter(
    # model, 'XGBoostXGBClassifier',
    # calculate_linear_classifier_output_shapes, convert_xgboost,
    # options={'nocl': [True, False], 'zipmap': [True, False, 'columns']})
    onx = convert_sklearn(model, initial_types=inputs_type, options={id(model): {'zipmap': False}})
    # onx = convert_sklearn(
    # model, 'pipeline_xgboost',
    # [('input', FloatTensorType([None, 2]))],
    # target_opset=12)

    with open(output_path, "wb") as f:
        f.write(onx.SerializeToString())



from setuptools import find_namespace_packages, setup, find_packages
import re


VERSION = open('VERSION').read()
LONG_DESCRIPTION = open('README.md').read()

_deps = [
    "numpy==1.21.5",
    "pyaml==21.10.1",
    "types-PyYAML==6.0.1",
    "onnxruntime==1.10.0",
    "onnxruntime-tools==1.7.0",
    "onnxmltools==1.10.0",
    "sympy==1.9.0",
    "unidecode==1.3.2",
    "torch==1.10.1",
    "transformers==4.15.0",
    "scikit-learn==1.0.1",
    "skl2onnx==1.9.0",
    "tensorflow==2.7.0",
    "tf2onnx==1.9.3",
    "xgboost==1.5.1",
    "pytest==6.2.5",
    "pytest-cov==3.0.0",
    "mock==4.0.3",
    "black==21.12b0",
    "flake8==4.0.1",
    "mypy==0.930",
    "isort==5.10.1"
]


deps = {b: a for a, b in (re.findall(r"^(([^!=<>]+)(?:[!=<>].*)?$)", x)[0] for x in _deps)}


def deps_list(*pkgs):
    return [deps[pkg] for pkg in pkgs]

extras = {}

extras["torch"] = deps_list("torch")
extras["tf"] = deps_list("tensorflow", "tf2onnx")
extras["transformers"] = (
    deps_list("transformers")
    + extras["torch"]
    + extras["tf"] 
)
extras["sklearn"] = deps_list("scikit-learn", "skl2onnx")
extras["xgboost"] = (
    deps_list("xgboost")
    + extras["sklearn"] 
)

extras["all"] = (
    extras["torch"]
    + extras["tf"]
    + extras["transformers"]
    + extras["sklearn"]
    + extras["xgboost"]
)

extras["dev"] = (
    deps_list(
        "pytest",
        "pytest-cov",
        "mock",
        "black",
        "flake8",
        "mypy",
        "isort"
    )
    + extras["all"] 
)

install_requires = [
    deps["numpy"],
    deps["pyaml"],
    deps["types-PyYAML"],
    deps["onnxruntime"],
    deps["onnxruntime-tools"],
    deps["onnxmltools"],
    deps["sympy"],
    deps["unidecode"]
]


setup(
    name='quick-deploy',
    version=VERSION,
    description='Quick-Deploy optimize and deploy Machine Learning models as fast inference API.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/rodrigobaron/quick-deploy",
    author="Rodrigo Baron",
    author_email="baron.rodrigo0@gmail.com",
    license="Apache",
    package_dir={'': 'src'},
    packages=find_packages(where="src"),
    package_data={
        'quick_deploy': ['py.typed'],
    },
    zip_safe=False,
    extras_require=extras,
    entry_points={
        'console_scripts': [
            'quick-deploy = quick_deploy.cli:main',
        ],
    },
    data_files=[('', ['VERSION', 'README.md', 'LICENSE'])],
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=install_requires,
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
